#!/usr/bin/env python3

# ----------------------------------------------------------------------------
# Copyright (c) 2020--, Qiyun Zhu.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

from unittest import TestCase, main
from os import remove
from os.path import join, dirname, realpath
from shutil import rmtree
from tempfile import mkdtemp
from filecmp import cmp
import gzip

from click.testing import CliRunner

from woltka.cli import (
    cli, classify_cmd, normalize_cmd, filter_cmd, merge_cmd, collapse_cmd,
    coverage_cmd)


class CliTests(TestCase):
    def setUp(self):
        self.tmpdir = mkdtemp()
        self.datdir = join(dirname(realpath(__file__)), 'data')
        self.alndir = join(self.datdir, 'align')
        self.taxdir = join(self.datdir, 'taxonomy')
        self.fundir = join(self.datdir, 'function')
        self.outdir = join(self.datdir, 'output')
        self.runner = CliRunner()

    def tearDown(self):
        rmtree(self.tmpdir)

    def test_cli(self):
        self.assertRaises(SystemExit, cli)

    def test_classify(self):
        output_fp = join(self.tmpdir, 'output.tsv')

        def _test_params(params, exp):
            res = self.runner.invoke(classify_cmd, params + ['--no-exe'])
            self.assertEqual(res.exit_code, 0)
            self.assertTrue(cmp(output_fp, join(self.outdir, exp)))

        # bowtie2, ogu
        params = ['--input',  join(self.alndir, 'bowtie2'),
                  '--output', output_fp]
        _test_params(params, 'bowtie2.ogu.tsv')

        # bowtie2, paired-end, free-rank classification
        params = ['--input',  join(self.alndir, 'bowtie2'),
                  '--output', output_fp,
                  '--nodes',  join(self.taxdir, 'nodes.dmp'),
                  '--map',    join(self.taxdir, 'taxid.map'),
                  '--rank',   'free']
        _test_params(params, 'bowtie2.free.tsv')

        # blastn, multiplexed, lineage-based, species-level classification
        params = ['--input',   join(self.alndir, 'blastn', 'mux.b6o.xz'),
                  '--output',  output_fp,
                  '--lineage', join(self.taxdir, 'lineages.txt'),
                  '--rank',    'species']
        _test_params(params, 'blastn.species.tsv')

        # burst, classification at genus level
        params = ['--input',  join(self.alndir, 'burst'),
                  '--output', output_fp,
                  '--outmap', self.tmpdir,
                  '--names',  join(self.taxdir, 'names.dmp'),
                  '--nodes',  join(self.taxdir, 'nodes.dmp'),
                  '--map',    join(self.taxdir, 'taxid.map'),
                  '--rank',   'genus',
                  '--name-as-id']
        _test_params(params, 'burst.genus.tsv')

        # check read maps
        for i in range(1, 6):
            outmap_fp = join(self.tmpdir, f'S0{i}.txt.gz')
            with gzip.open(outmap_fp, 'r') as f:
                obs = f.read()
            with gzip.open(join(self.outdir, 'burst.genus.map',
                                f'S0{i}.txt.gz'), 'r') as f:
                exp = f.read()
            self.assertEqual(obs, exp)
            remove(outmap_fp)

        # blastn, family-level classification, report relative abundance (%)
        params = ['--input',  join(self.alndir, 'blastn', 'mux.b6o.xz'),
                  '--output', output_fp,
                  '--names',  join(self.taxdir, 'names.dmp'),
                  '--nodes',  join(self.taxdir, 'nodes.dmp'),
                  '--map',    join(self.taxdir, 'taxid.map'),
                  '--rank',   'family',
                  '--name-as-id',
                  '--frac',
                  '--scale',  '100',
                  '--digits', 2]
        _test_params(params, 'blastn.family.percent.tsv')

        # bt2sho, order-level, genome size-normalized classification
        # unit: reads per million bases (of genome)
        params = ['--input',  join(self.alndir, 'bt2sho'),
                  '--output', output_fp,
                  '--names',  join(self.taxdir, 'names.dmp'),
                  '--nodes',  join(self.taxdir, 'nodes.dmp'),
                  '--map',    join(self.taxdir, 'taxid.map'),
                  '--rank',   'order',
                  '--sizes',  join(self.taxdir, 'length.map'),
                  '--scale',  '1M',
                  '--digits', 3]
        _test_params(params, 'bt2sho.order.cpm.tsv')

        # bt2sho phylogeny-based classification
        params = ['--input',  join(self.alndir, 'bt2sho'),
                  '--output', output_fp,
                  '--newick', join(self.datdir, 'tree.nwk'),
                  '--rank',   'free',
                  '--subok']
        _test_params(params, 'bt2sho.phylo.tsv')

        # burst, classification by GO process
        params = ['--input',  join(self.alndir, 'burst'),
                  '--output', output_fp,
                  '--rank',   'process',
                  '--coords', join(self.fundir, 'coords.txt.xz'),
                  '--map',    join(self.fundir, 'uniref', 'uniref.map.xz'),
                  '--map',    join(self.fundir, 'go', 'process.tsv.xz')]
        _test_params(params, 'burst.process.tsv')

        # burst, stratified genus/process classification
        params = ['--input',  join(self.alndir, 'burst'),
                  '--output', output_fp,
                  '--rank',   'process',
                  '--coords', join(self.fundir, 'coords.txt.xz'),
                  '--map',    join(self.fundir, 'uniref', 'uniref.map.xz'),
                  '--map',    join(self.fundir, 'go', 'process.tsv.xz'),
                  '--stratify', join(self.outdir, 'burst.genus.map')]
        _test_params(params, 'burst.genus.process.tsv')

        # bt2sho, classification by GO component, normalization by gene length
        # unit: reads per kilobase (RPK)
        params = ['--input',  join(self.alndir, 'bt2sho'),
                  '--output', output_fp,
                  '--rank',   'component',
                  '--coords', join(self.fundir, 'coords.txt.xz'),
                  '--map',    join(self.fundir, 'uniref', 'uniref.map.xz'),
                  '--map',    join(self.fundir, 'go', 'component.tsv.xz'),
                  '--sizes',  '.',
                  '--scale',  '1k',
                  '--digits', 3]
        _test_params(params, 'bt2sho.component.rpk.tsv')

        # simple map (from burst) against genes, classification at genus level
        params = ['--input',    join(self.alndir, 'burst', 'split'),
                  '--trim-sub', '_',
                  '--output',   output_fp,
                  '--rank',     'genus',
                  '--map',      join(self.taxdir, 'nucl', 'nucl2tid.txt'),
                  '--names',    join(self.taxdir, 'names.dmp'),
                  '--nodes',    join(self.taxdir, 'nodes.dmp'),
                  '--name-as-id']
        _test_params(params, 'split.genus.tsv')

        # simple map against genes, classification by GO process
        params = ['--input',  join(self.alndir, 'burst', 'split'),
                  '--output', output_fp,
                  '--rank',   'process',
                  '--map',    join(self.fundir, 'nucl', 'uniref.map.xz'),
                  '--map',    join(self.fundir, 'go', 'process.tsv.xz')]
        _test_params(params, 'split.process.tsv')

        remove(output_fp)

    def test_classify_cmd(self):
        input_fp = join(self.datdir, 'output', 'bowtie2.ogu.tsv')
        output_fp = join(self.tmpdir, 'tmp.tsv')
        sizes_fp = join(self.datdir, 'taxonomy', 'length.map')
        params = ['--input',  input_fp,
                  '--output', output_fp,
                  '--sizes',  sizes_fp,
                  '--scale',  '1M']
        res = self.runner.invoke(normalize_cmd, params)
        self.assertEqual(res.exit_code, 0)
        self.assertEqual(res.output.splitlines()[-1],
                         'Normalized profile written.')
        with open(output_fp, 'r') as f:
            obs = f.read().splitlines()
        self.assertIn('G000007325\t0\t12\t0\t0\t104', obs)
        self.assertIn('G000215745\t0\t0\t0\t358\t0', obs)
        remove(output_fp)

    def test_filter_cmd(self):
        input_fp = join(self.outdir, 'bowtie2.free.tsv')
        output_fp = join(self.tmpdir, 'output.tsv')
        params = ['--input',  input_fp,
                  '--output', output_fp,
                  '--min-percent', 1]
        res = self.runner.invoke(filter_cmd, params)
        self.assertEqual(res.exit_code, 0)
        self.assertEqual(res.output.splitlines()[-1],
                         'Filtered profile written.')
        with open(output_fp, 'r') as f:
            self.assertEqual(len(f.read().splitlines()), 30)
        remove(output_fp)

    def test_merge_cmd(self):
        input_fp1 = join(self.outdir, 'burst.process.tsv')
        input_fp2 = join(self.outdir, 'split.process.tsv')
        output_fp = join(self.tmpdir, 'output.tsv')
        params = ['--input',  input_fp1,
                  '--input',  input_fp2,
                  '--output', output_fp]
        res = self.runner.invoke(merge_cmd, params)
        self.assertEqual(res.exit_code, 0)
        self.assertEqual(res.output.splitlines()[-1],
                         'Merged profile written.')
        with open(output_fp, 'r') as f:
            self.assertEqual(len(f.read().splitlines()), 205)
        remove(output_fp)

    def test_collapse_cmd(self):
        input_fp = join(self.outdir, 'truth.uniref.tsv')
        map_fp = join(self.fundir, 'go', 'goslim.tsv.xz')
        output_fp = join(self.tmpdir, 'output.tsv')
        names_fp = join(self.fundir, 'go', 'name.txt.xz')
        params = ['--input',  input_fp,
                  '--map',    map_fp,
                  '--output', output_fp,
                  '--names',  names_fp,
                  '--divide']
        res = self.runner.invoke(collapse_cmd, params)
        self.assertEqual(res.exit_code, 0)
        self.assertEqual(res.output.splitlines()[-1],
                         'Collapsed profile written.')
        with open(output_fp, 'r') as f:
            self.assertEqual(len(f.read().splitlines()), 74)
        remove(output_fp)

    def test_coverage_cmd(self):
        input_fp = join(self.outdir, 'truth.metacyc.tsv')
        map_fp = join(self.fundir, 'metacyc', 'pathway_mbrs.txt')
        output_fp = join(self.tmpdir, 'output.tsv')
        names_fp = join(self.fundir, 'metacyc', 'pathway_name.txt')
        params = ['--input',  input_fp,
                  '--map',    map_fp,
                  '--output', output_fp,
                  '--names',  names_fp]
        res = self.runner.invoke(coverage_cmd, params)
        self.assertEqual(res.exit_code, 0)
        self.assertEqual(res.output.splitlines()[-1],
                         'Coverage table written.')
        with open(output_fp, 'r') as f:
            self.assertEqual(len(f.read().splitlines()), 316)
        remove(output_fp)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3

# ----------------------------------------------------------------------------
# Copyright (c) 2020--, Qiyun Zhu.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

"""Functions for handling hierarchical classification systems.

Notes
-----
A hierarchical classification system is represented by a tree, which is a dict
of taxon ID to parent taxon ID. Each taxon ID is a unique identifier (such as
NCBI TaxID), although a descriptive name (e.g., "Escherichia coli") may work as
long as it is unique. The root of the tree is indicated by ID == parent ID. The
terminal leaves of the tree are IDs of subjects to which input data are mapped.
This simple data structure is sufficient for the purpose of this program.

In addition, the name and rank assignments of taxa are recorded in two separate
dicts, if available. In principle, the system is rank-independent (i.e., "rank"
can be anything or left empty).
"""

import re

from .util import last_value


# standard taxonomic ranks
rankorder = ('kingdom', 'phylum', 'class', 'order', 'family', 'genus',
             'species', 'strain')
rank2code = {'kingdom': 'k', 'phylum': 'p', 'class':   'c', 'order':  'o',
             'family':  'f', 'genus':  'g', 'species': 's', 'strain': 't'}
code2rank = {v: k for k, v in rank2code.items()}

# compatibilities for top-level ranks (domain, superkingdom, kingdom)
rank2code['domain'] = 'k'
rank2code['superkingdom'] = 'k'
code2rank['d'] = 'kingdom'

# unclassified taxon names
notax = {'', '0', 'unclassified', 'unassigned'}


def read_names(fh):
    """Read taxonomic names from a file.

    Parameters
    ----------
    fh : file handle
        Taxon name mapping file.

    Returns
    -------
    dict
        Taxon name map.

    Notes
    -----
    Can be NCBI-style names.dmp or a plain map of ID to name.
    """
    names = {}
    for line in fh:
        x = line.rstrip().replace('\t|', '').split('\t')
        if len(x) < 4 or x[3] == 'scientific name':
            names[x[0]] = x[1]
    return names


def read_nodes(fh):
    """Read taxonomic nodes from a file.

    Parameters
    ----------
    fh : file handle
        Taxonomic nodes file.

    Returns
    -------
    dict
        Taxonomy tree.
    dict
        Taxonomic rank map.

    Notes
    -----
    Input file can be NCBI-style nodes.dmp or a plain map of ID to parent and
    (optional) rank.
    """
    tree, rankdic = {}, {}
    for line in fh:
        x = line.rstrip().replace('\t|', '').split('\t')
        tree[x[0]] = x[1]
        try:
            rankdic[x[0]] = x[2]
        except IndexError:
            pass
    return tree, rankdic


def read_newick(fh):
    """Read tree from a Newick-format file.

    Parameters
    ----------
    fh : file handle
        Newick tree file.

    Returns
    -------
    dict
        Taxonomy tree.

    Raises
    ------
    ValueError
        Missing internal node ID.
    ValueError
        Found non-unique node ID.

    Notes
    -----
    This simple Newick parser only captures one piece of information: child-to-
    parent mapping. All others (such as branch lengths) are omitted. It assumes
    that the Newick format is correct, and only raises at two scenarios where
    the tree does not satisfy the need of this program.
    """
    # get clean Newick string
    nwk = ''.join(x.strip() for x in fh).rstrip(';')
    res = {}

    # Newick node ("(a,b)")
    pnod = re.compile(r'\([^()]+\)')

    # Newick node end ("," or ")")
    pend = re.compile(r'[,)]')

    # Newick node label ("Id:length")
    def _get_id(label):
        return label.split(':', 1)[0].strip('"\'')

    # recursively search for nodes from inside to outside
    while True:
        m = pnod.search(nwk)
        if m is None:
            break

        # get parent Id of current node
        tail = nwk[m.end(0):]
        parent = _get_id(pend.split(tail, 1)[0])
        if parent == '':
            raise ValueError('Missing internal node ID.')

        # get child Ids of current node
        for child in [_get_id(x) for x in m.group(0)[1:-1].split(',')]:
            if child in res:
                raise ValueError(f'Found non-unique node ID: "{child}".')
            res[child] = parent

        # remove node from search string
        nwk = nwk[:m.start(0)] + tail

    # outest-most parent must be root
    res[parent] = parent
    return res


def read_columns(fh):
    """Read taxonomic information from a rank-per-column table.

    Parameters
    ----------
    fh : file handle
        Rank table file.

    Returns
    -------
    dict
        Taxonomy tree.
    dict
        Rank dictionary.

    Notes
    -----
    Taxonomic units must be unique.

    TODO
    ----
    Add support for same taxon name but different ranks. This may be useful.
    For example, "Actinobacteria" is both a phylum and a class.
    """
    tree, rankdic = {}, {}

    # get rank names from header
    ranks = next(fh).rstrip().split('\t')[1:]

    # ranks = None
    for line in fh:
        row = line.rstrip().split('\t')

        # get lineage (taxa from high to low)
        lineage = [None if x in notax else x for x in row[1:]]

        # map entry to lowest taxon
        tree[row[0]] = last_value(lineage)

        for i, taxon in enumerate(lineage):
            if taxon is None:
                continue

            # map lower taxon to higher taxon
            rank, parent = ranks[i], last_value(lineage[:i])

            # check existing relationship
            try:
                if tree[taxon] != parent or rankdic[taxon] != rank:
                    raise ValueError(f'Conflict at taxon "{taxon}".')

            # add current taxon to dictionary
            except KeyError:
                tree[taxon], rankdic[taxon] = parent, rank

    return tree, rankdic


def read_lineage(fh):
    """Read taxonomic information from a lineage mapping file.

    Parameters
    ----------
    fh : file handle
        Taxonomic lineage mapping file.

    Returns
    -------
    dict
        Taxonomy tree.
    dict
        Rank dictionary.

    Raises
    ------
    ValueError
        Conflict of taxon-parent relationship is found.

    Notes
    -----
    The lineage mapping file format is a.k.a. Greengenes-style.

    A lineage string consists of one or multiple semicolon-delimited taxa.
    Spaces leading or trailing a taxon are stripped. Spaces within a taxon
    are tolerable.

    Each taxon is represented by the entire ancestral lineage before it.
    Therefore, the taxon itself does not have to be unique, but the ancestral
    lineage does.

    Empty levels in the end are discarded. e.g., "k__Bacteria;p__" is not a
    valid taxon.

    Empty levels in the middle are kept. e.g., "k__Bacteria;p__;c_Clostridia"
    will not be shortened into "k__Bacteria;c_Clostridia".
    """
    p = re.compile(r'([a-z])__.*')
    tree, rankdic = {}, {}
    for line in fh:
        if line.startswith('#'):
            continue
        id_, lineage = line.rstrip().split('\t')
        parent, this = None, None
        for taxon in lineage.split(';'):
            taxon = taxon.strip()

            # build the entire lineage
            this = f'{this};{taxon}' if this else taxon

            # skip empty taxon
            if taxon.lower() in notax or taxon[1:] == '__':
                continue

            # add current taxon to dictionary
            tree[this] = parent

            # get rank from prefix (e.g., "k__")
            m = p.match(taxon)
            if m:
                rank = m.group(1)
                if rank in code2rank:
                    rankdic[this] = code2rank[rank]

            parent = this

        # add entry to whole lineage
        tree[id_] = parent

    return tree, rankdic


def fill_root(tree):
    """Add a root node to a tree if absent.

    Parameters
    ----------
    tree : dict
        Taxonomy tree.

    Returns
    -------
    str
        Root node identifier

    Notes
    -----
    A root is defined as having parent as itself, a behavior derived from the
    NCBI convention. Exactly one root must be present in a tree, so that all
    taxa can be traced back to the same root.

    In custom trees, there may or may not be a clearly defined root. This
    function aims as defining a root for any given tree. Specifically, if
    there is one root node (highest hierarchy), this node will be "sealed"
    (making parent as itself). If there are multiple crown nodes ("crown"
    describes the root of a clade instead of the entire tree), a new root
    node will be generated and serve as the parent of all crown nodes.
    """
    crown, toadd, tested = [], set(), set()
    for taxon in tree:
        this = taxon
        while True:
            if this in tested:
                break
            tested.add(this)

            # parent is missing
            try:
                parent = tree[this]
            except KeyError:
                crown.append(this)
                toadd.add(this)
                break

            # parent is None
            if parent is None:
                crown.append(this)
                break

            # parent is itself
            if parent == this:
                crown.append(this)
                break
            this = parent

    # fill non-existent root or crown nodes
    for node in toadd:
        tree[node] = None

    # this happens only when tree is empty
    if len(crown) == 0:
        return None

    # there is only one crown node
    elif len(crown) == 1:

        # make the parent of root iself
        root = crown[0]
        tree[root] = root
        return root

    # there are more than one crown node
    else:

        # in NCBI convention, root should have identifier "1"
        i = 1

        # in case "1" is already in tree, find an unused integer
        while True:
            if str(i) not in tree:
                break
            i += 1
        root = str(i)
        tree[root] = root

        # coalesce all crown nodes to root
        for x in crown:
            tree[x] = root
        return root


def get_lineage(taxon, tree):
    """Get lineage of given taxon in taxonomy tree.

    Parameters
    ----------
    taxon : str
        Query taxon.
    tree : dict
        Taxonomy tree.

    Returns
    -------
    list of str
        Lineage from root to query taxon.
    """
    # if taxon is not in tree, return None
    try:
        parent = tree[taxon]
    except KeyError:
        return None

    # initiate lineage and cache method reference
    lineage = [taxon]
    add_taxon = lineage.append

    # move up classification hierarchy till root
    this = taxon
    while True:

        # stop when reaching root
        if parent == this:
            break

        # add current level to lineage
        add_taxon(parent)

        # move up to parent
        this = parent
        parent = tree[this]

    # make lineage high-to-low
    return lineage[::-1]


def lineage_str(taxon, tree, namedic=None, include_self=False,
                include_root=False):
    """Generate a Greengenes-style lineage string of a taxon.

    Parameters
    ----------
    taxon : str
        Query taxon.
    tree : dict
        Taxonomy tree.
    namedic : dict, optional
        Taxon name dictionary.
    include_self : bool, optional
        Include self.
    include_root : bool, optional
        Include root.

    Returns
    -------
    list of str
        Lineage from root to query taxon.
    """
    lineage = get_lineage(taxon, tree)
    if lineage is None:
        return ''
    start = 0 if include_root else 1
    end = len(lineage) if include_self else len(lineage) - 1
    taxa = lineage[start:end]
    return ';'.join([
        namedic[x] if namedic and x in namedic else x for x in taxa])


def find_rank(taxon, rank, tree, rankdic):
    """Find ancestor at a given rank by stepping up the classification
    hierarchy.

    Parameters
    ----------
    taxon : str
        Query taxon.
    rank : str
        Target rank.
    tree : dict
        Taxonomy tree.
    rankdic : dict
        Taxon-to-rank map.

    Returns
    -------
    str or None
        Ancestral taxon at given rank, or None if not found.
    """
    # if taxon is not in tree, return None
    try:
        parent = tree[taxon]
    except KeyError:
        return

    # cache method reference
    get_rank = rankdic.get

    # move up hierarchy until reaching given rank
    this = taxon
    while True:

        # check rank of current taxon
        if get_rank(this) == rank:
            return this

        # stop when reaching root
        if parent == this:
            break

        # move up to parent
        this = parent
        parent = tree[this]


def find_lca(taxa, tree):
    """Find lowest common ancestor (LCA) of given taxa.

    Parameters
    ----------
    taxa : iterable of str
        Query taxa.
    tree : dict
        Taxonomy tree.

    Returns
    -------
    str or None
        LCA of taxa, or None if any taxon is not in tree.

    TODO
    ----
    Combine LCA and majority rule, which is not trivial and requires careful
    reasoning and algorithm design.
    """
    itaxa = iter(taxa)

    # get lineage of first taxon
    lineage = get_lineage(next(itaxa), tree)
    if lineage is None:
        return

    # compare with remaining taxa
    for taxon in itaxa:
        try:
            parent = tree[taxon]
        except KeyError:
            return
        this = taxon
        while True:

            # look for shared lineage
            try:
                idx = lineage.index(this)

            # if not found
            except ValueError:
                if parent == this:
                    break
                this = parent
                parent = tree[this]

            # trim shared lineage
            else:
                if idx + 1 < len(lineage):
                    lineage = lineage[slice(idx + 1)]
                break

    # LCA is the last of shared lineage
    return lineage[-1]



#!/usr/bin/env python3

# ----------------------------------------------------------------------------
# Copyright (c) 2020--, Qiyun Zhu.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

from unittest import TestCase, main
from os.path import join, dirname, realpath
from shutil import rmtree
from tempfile import mkdtemp

from woltka.util import (
    add_dict, update_dict, sum_dict, scale_factor, scale_dict, intize,
    intize_list, intize_dict, rounder, round_list, round_dict, delnone,
    allkeys, count_list, last_value, feature_count)


class UtilTests(TestCase):
    def setUp(self):
        self.tmpdir = mkdtemp()
        self.datdir = join(dirname(realpath(__file__)), 'data')

    def tearDown(self):
        rmtree(self.tmpdir)

    def test_add_dict(self):
        d = {'a': 1, 'b': 2}
        add_dict(d, 'c', 3)
        self.assertDictEqual(d, {'a': 1, 'b': 2, 'c': 3})
        add_dict(d, 'b', 2)
        self.assertDictEqual(d, {'a': 1, 'b': 2, 'c': 3})
        with self.assertRaises(AssertionError) as ctx:
            add_dict(d, 'a', 0)
        self.assertEqual(str(ctx.exception), (
            'Conflicting values found for "a".'))

    def test_update_dict(self):
        d0 = {'a': 1, 'b': 2}
        d1 = {'c': 3, 'd': 4}
        update_dict(d0, d1)
        self.assertDictEqual(d0, {'a': 1, 'b': 2, 'c': 3, 'd': 4})
        d2 = {'e': 5, 'a': 1}
        update_dict(d0, d2)
        self.assertDictEqual(d0, {'a': 1, 'b': 2, 'c': 3, 'd': 4, 'e': 5})
        d3 = {'g': 6, 'b': 100}
        with self.assertRaises(AssertionError) as ctx:
            update_dict(d0, d3)
        self.assertEqual(str(ctx.exception), (
            'Conflicting values found for "b".'))

    def test_sum_dict(self):
        d0 = {'a': 1, 'b': 2}
        d1 = {'c': 3, 'd': 4}
        sum_dict(d0, d1)
        self.assertDictEqual(d0, {'a': 1, 'b': 2, 'c': 3, 'd': 4})
        d2 = {'a': 1}
        sum_dict(d0, d2)
        self.assertDictEqual(d0, {'a': 2, 'b': 2, 'c': 3, 'd': 4})
        d3 = {'e': 5, 'b': 3}
        sum_dict(d0, d3)
        self.assertDictEqual(d0, {'a': 2, 'b': 5, 'c': 3, 'd': 4, 'e': 5})

    def test_scale_factor(self):
        self.assertEqual(scale_factor('100'), 100)
        self.assertEqual(scale_factor('0.05'), 0.05)
        self.assertEqual(scale_factor('1k'), 1000)
        self.assertEqual(scale_factor('2M'), 2000000)
        self.assertEqual(scale_factor('1e9'), 1.0e+9)
        with self.assertRaises(ValueError) as ctx:
            scale_factor('not_a_number')
        self.assertEqual(str(ctx.exception), (
            'Invalid scale factor.'))

    def test_scale_dict(self):
        d = {'a': 1.0, 'b': 1.5, 'c': 2.4}
        scale_dict(d, 2)
        exp = {'a': 2.0, 'b': 3.0, 'c': 4.8}
        self.assertDictEqual(d, exp)
        scale_dict(d, 100)
        exp = {'a': 200, 'b': 300, 'c': 480}
        self.assertDictEqual(d, exp)

    def test_intize(self):
        self.assertEqual(intize(1), 1)
        self.assertEqual(intize(1.0), 1)
        self.assertEqual(intize(1.1), 1)
        self.assertEqual(intize(1.49), 1)
        self.assertEqual(intize(1.49999), 1)
        self.assertEqual(intize(1.49999999999), 2)
        self.assertEqual(intize(1.5), 2)
        self.assertEqual(intize(1.8), 2)
        self.assertEqual(intize(2.2), 2)
        self.assertEqual(intize(2.5), 2)
        self.assertEqual(intize(2.50001), 3)
        self.assertEqual(intize(2.50000000001), 2)
        self.assertEqual(intize(2.7), 3)
        self.assertEqual(intize(-1.0), -1)
        self.assertEqual(intize(-1.5), -2)
        self.assertEqual(intize(-1.49999999999), -2)

    def test_intize_list(self):
        lst = [0.1, 1.2, 2.4, 3.8, 4.5, 5.5]
        intize_list(lst)
        exp = [0, 1, 2, 4, 4, 6]
        self.assertListEqual(lst, exp)

    def test_intize_dict(self):
        dic = {'a': 1.0, 'b': 2.2, 'c': 3.6}
        intize_dict(dic)
        exp = {'a': 1, 'b': 2, 'c': 4}
        self.assertDictEqual(dic, exp)
        dic = {'a': -0.2, 'b': -3.3, 'c': 1.8, 'd': 0.4}
        intize_dict(dic)
        exp = {'b': -3, 'c': 2}
        self.assertDictEqual(dic, exp)
        dic = {'a': -0.2, 'b': -3.3, 'c': 1.8, 'd': 0.4}
        intize_dict(dic, zero=True)
        exp = {'a': 0, 'b': -3, 'c': 2, 'd': 0}
        self.assertDictEqual(dic, exp)
        dic = {'a': 2.5, 'b': 3.5, 'c': 1.49999999999, 'd': 1.50000000001}
        intize_dict(dic)
        exp = {'a': 2, 'b': 4, 'c': 2, 'd': 2}
        self.assertDictEqual(dic, exp)

    def test_rounder(self):
        self.assertEqual(rounder(1.0), 1)
        self.assertEqual(rounder(1.49), 1)
        self.assertEqual(rounder(1.49999999999), 2)
        self.assertEqual(rounder(1.5), 2)
        self.assertEqual(rounder(0.02, 1), 0.0)
        self.assertEqual(rounder(1.05, 1), 1.1)
        self.assertEqual(rounder(1.049, 1), 1.0)
        self.assertEqual(rounder(1.745, 2), 1.75)
        self.assertEqual(rounder(4.235, 2), 4.24)
        self.assertEqual(rounder(4.2350000000001, 2), 4.24)
        self.assertEqual(rounder(1.74499999, 2), 1.74)
        self.assertEqual(rounder(1.7449999999999, 2), 1.75)

    def test_round_list(self):
        lst = [0.1, 1.2, 2.4, 3.8, 4.5, 5.5]
        round_list(lst)
        exp = [0, 1, 2, 4, 4, 6]
        self.assertListEqual(lst, exp)
        lst = [1.71828, 3.14159, 2.00005, 3.874999, 4.500001]
        round_list(lst, 2)
        exp = [1.72, 3.14, 2.0, 3.87, 4.5]
        self.assertListEqual(lst, exp)

    def test_round_dict(self):
        dic = {'a': 1.0, 'b': 2.2, 'c': 3.6}
        round_dict(dic)
        exp = {'a': 1, 'b': 2, 'c': 4}
        self.assertDictEqual(dic, exp)
        dic = {'a': 1.05, 'b': 2.24, 'c': 3.68, 'd': 0.02}
        round_dict(dic, 1)
        exp = {'a': 1.1, 'b': 2.2, 'c': 3.7}
        self.assertDictEqual(dic, exp)
        dic = {'a': 2.225, 'b': 3.505, 'c': 1.7449999999999,
               'd': 4.2350000000001}
        round_dict(dic, 2)
        exp = {'a': 2.23, 'b': 3.5, 'c': 1.75, 'd': 4.24}
        self.assertDictEqual(dic, exp)
        dic = {'a': -0.24, 'b': 3.71, 'c': 0.02}
        round_dict(dic, digits=1, zero=True)
        exp = {'a': -0.2, 'b': 3.7, 'c': 0.0}
        self.assertDictEqual(dic, exp)

    def test_delnone(self):
        dic = {'a': 1, 'b': 2, 'c': 3}
        obs = dic.copy()
        delnone(obs)
        self.assertDictEqual(obs, dic)
        obs[None] = 4
        delnone(obs)
        self.assertDictEqual(obs, dic)

    def test_allkeys(self):
        data = {'S1': {'G1': 4, 'G2': 5, 'G3': 8},
                'S2': {'G1': 2, 'G4': 3, 'G5': 7},
                'S3': {'G2': 3, 'G5': 5}}
        obs = allkeys(data)
        exp = {'G1', 'G2', 'G3', 'G4', 'G5'}
        self.assertSetEqual(obs, exp)

    def test_count_list(self):
        lst = [1, 1, 2, 3, 1, 2, 4, 3, 1]
        obs = count_list(lst)
        exp = {1: 4, 2: 2, 3: 2, 4: 1}
        self.assertDictEqual(obs, exp)

    def test_last_value(self):
        lst = ['a', 1, None, 'b', 2, None]
        self.assertEqual(last_value(lst), 2)
        self.assertIsNone(last_value([None, None]))

    def test_feature_count(self):
        self.assertTupleEqual(feature_count('A'),   ('A', 1))
        self.assertTupleEqual(feature_count('A:1'), ('A', 1))
        self.assertTupleEqual(feature_count('A:2'), ('A', 2))
        self.assertTupleEqual(feature_count('A:0'), ('A:0', 1))
        self.assertTupleEqual(feature_count('A:x'), ('A:x', 1))
        self.assertTupleEqual(feature_count(':1'),  (':1', 1))
        self.assertTupleEqual(feature_count('::'),  ('::', 1))
        self.assertTupleEqual(feature_count(''),    ('', 1))


if __name__ == '__main__':
    main()



#!/usr/bin/env python3

# ----------------------------------------------------------------------------
# Copyright (c) 2020--, Qiyun Zhu.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

"""Functions for classifying query sequences by assigning their subjects to a
hierarchical classification system.
"""

from operator import itemgetter
from collections import defaultdict

from .util import count_list
from .tree import find_rank, find_lca


"""
Assigner: assign a query sequence to one or more taxa based on its subjects and
a classification system, if applicable.

- assign_none
- assign_free
- assign_rank
"""


def assign_none(subs, uniq=False):
    """Assign query to subjects without using a classification system.

    Parameters
    ----------
    subs : tuple of str
        Subjects.
    uniq : bool, optional
        Assignment must be unique.

    Returns
    -------
    str or list or None
        Unique subject or list of subjects.
    """
    try:
        sub, = subs
        return sub
    except ValueError:
        return None if uniq else list(subs)


def assign_free(subs, tree, root=None, subok=False):
    """Assign query based on a classification system in a rank-free manner.

    Parameters
    ----------
    subs : tuple of str
        Subjects.
    tree : dict
        Hierarchical classification system.
    root : str, optional
        Root identifier.
    subok : bool, optional
        Allow assignment to subjects.

    Returns
    -------
    str or None
        Unique assignment.
    """
    try:
        sub, = subs
        return sub if subok else (tree[sub] if sub in tree else None)
    except ValueError:
        lca = find_lca(subs, tree)
        return None if lca == root else lca


def assign_rank(subs, rank, tree, rankdic, root=None, major=None, above=False,
                uniq=False):
    """Assign query to a given rank in a classification system.

    Parameters
    ----------
    subs : tuple of str
        Subjects.
    rank : str
        Target rank.
    tree : dict
        Hierarchical classification system.
    rankdic : dict
        Rank dictionary.
    root : str, optional
        Root identifier.
    major : float, optional
        Majority-rule assignment threshold.
    above : bool, optional
        Allow assignment above rank.
    uniq : bool, optional
        Assignment must be unique.

    Returns
    -------
    str or list or None
        Unique assignment or list of assignments.

    TODO
    ----
    Combine major and above.
    """
    taxa = [find_rank(x, rank, tree, rankdic) for x in subs]
    tset = set(taxa)
    if len(tset) == 1:
        return taxa[0]
    elif major:
        return majority(taxa, major)
    elif above:
        if None in tset:
            return None
        lca = find_lca(tset, tree)
        return None if lca == root else lca
    elif uniq:
        return None
    else:
        return taxa


"""
Counter: convert a read-to-taxa map into a taxon-to-count map.

Four counter functions are implemented to tackle alternative parameter settings
(normalization by subject size and stratification by query), while maintaining
performance by saving conditional statements.

- counter
- counter_size
- counter_strat
- counter_size_strat
"""


def counter(taxque):
    """Count occurrences of taxa in a read-to-taxa map.

    Parameters
    ----------
    taxque : iterable of str or list

    Returns
    -------
    dict of int
        Map of taxon to count.
    """
    res = defaultdict(int)
    for taxa in taxque:
        if not taxa:
            continue

        # taxa is string: unique match
        try:
            res[taxa] += 1

        # taxa is list: multiple matches, to be divided by match count
        except TypeError:
            taxa = list(filter(None, taxa))
            k = 1 / len(taxa)
            for taxon in taxa:
                res[taxon] += k
    return res


def counter_size(subque, taxque, sizes):
    """Count occurrences of taxa in a read map while normalizing by subject
    size.

    Parameters
    ----------
    subque : iterable of tuple
        Subject(s) mapped to each query.
    taxque : iterable of str or list
        Taxon(a) assigned to each query.
    sizes : dict
        Subject size dictionary.

    Returns
    -------
    dict of float
        Map of taxon to normalized count.

    Raises
    ------
    KeyError
        Subject not found in size dictionary.

    See Also
    --------
    counter
    """
    res = defaultdict(int)
    for subs, taxa in zip(subque, taxque):
        if not taxa:
            continue
        try:
            res[taxa] += sum(sizes[x] for x in subs) / len(subs)
        except TypeError:
            k = 1 / len(list(filter(None, taxa)))
            for taxon, sub in zip(taxa, subs):
                if not taxon:
                    continue
                res[taxon] += sizes[sub] * k
    return res


def counter_strat(qryque, taxque, strata):
    """Stratify taxa in a read map and count occurrences.

    Parameters
    ----------
    qryque : iterable of str
        Query sequences.
    taxque : iterable of str or list
        Taxon(a) assigned to each query.
    strata : dict
        Query-to-stratum map.

    Returns
    -------
    dict of int
        Map of (stratum, taxon) to count.

    See Also
    --------
    counter
    """
    res = defaultdict(int)
    for query, taxa in zip(qryque, taxque):
        if not taxa or query not in strata:
            continue
        stratum = strata[query]
        try:
            res[(stratum, taxa)] += 1
        except TypeError:
            taxa = list(filter(None, taxa))
            k = 1 / len(taxa)
            for taxon in taxa:
                res[(stratum, taxon)] += k
    return res


def counter_size_strat(qryque, subque, taxque, sizes, strata):
    """Stratify taxa in a read map and count occurrences while normalizing by
    subject size.

    Parameters
    ----------
    qryque : iterable of str
        Query sequences.
    subque : iterable of tuple
        Subject(s) mapped to each query.
    taxque : iterable of str or list
        Taxon(a) assigned to each query.
    sizes : dict, optional
        Subject size dictionary.
    strata : dict
        Query-to-stratum map.

    Returns
    -------
    dict of float
        Map of (stratum, taxon) to normalized count.

    Raises
    ------
    KeyError
        Subject not found in size dictionary.

    See Also
    --------
    counter_size
    counter_strat
    """
    res = defaultdict(int)
    for query, subs, taxa in zip(qryque, subque, taxque):
        if not taxa or query not in strata:
            continue
        stratum = strata[query]
        try:
            res[(stratum, taxa)] += sum(sizes[x] for x in subs) / len(subs)
        except TypeError:
            k = 1 / len(list(filter(None, taxa)))
            for taxon, sub in zip(taxa, subs):
                if not taxon:
                    continue
                res[(stratum, taxon)] += sizes[sub] * k
    return res


def majority(taxa, th=0.8):
    """Select taxon from list by majority rule.

    Parameters
    ----------
    taxa : list of str
        Input taxon list.
    th : float, optional
        Threshold of majority, range = (0.5, 1.0].

    Returns
    -------
    str or None
        Selected taxon.
    """
    for taxon, n in sorted(count_list(taxa).items(), key=itemgetter(1),
                           reverse=True):
        return taxon if n >= len(taxa) * th else None



#!/usr/bin/env python3

# ----------------------------------------------------------------------------
# Copyright (c) 2020--, Qiyun Zhu.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------


"""Functions for parsing alignment / mapping files.

Notes
-----
A parser function operates on a single line in the input file.
A parser function returns a tuple of:
    str : query Id
    str : subject Id
    float : alignment score, optional
    int : alignment length, optional
    start : alignment start (5') coordinate, optional
    start : alignment end (3') coordinate, optional
"""

from collections import deque
from itertools import chain
from functools import lru_cache


def plain_mapper(fh, fmt=None, n=1000):
    """Read an alignment file in chunks and yield query-to-subject(s) maps.

    Parameters
    ----------
    fh : file handle
        Alignment file to parse.
    fmt : str, optional
        Alignment file format.
    n : int, optional
        Number of lines per chunk.

    Yields
    ------
    deque of str
        Query queue.
    deque of set of str
        Subject(s) queue.

    Notes
    -----
    The design of this function aims to couple with the extremely large size of
    typical alignment files. It reads the entire file sequentially, pauses and
    processes current cache for every _n_ lines, yields and clears cache, then
    proceeds.
    """
    # determine alignment file format
    fmt, head = (fmt, []) if fmt else infer_align_format(fh)

    # assign parser for given format
    parser = assign_parser(fmt)

    # query queue and subject(s) queue
    qryque, subque = deque(), deque()

    # pre-load method references
    qry_append, sub_append = qryque.append, subque.append

    this = None  # current query Id
    target = n   # target line number at end of current chunk

    # parse alignment file
    for i, (query, subject) in enumerate(parser(chain(iter(head), fh))):

        # add subject to subject set of the same query Id
        if query == this:
            subque[-1].add(subject)

        # when query Id changes,
        else:

            # line number has reached target
            if i >= target:

                # flush current queries and subject sets
                # this happens only when: 1) query Id changes, and 2) line
                # number has reached target, so that subjects of the same query
                # won't be separated in multiple flushes
                yield qryque, subque

                # re-initiate queues and method references
                qryque, subque = deque(), deque()
                qry_append, sub_append = qryque.append, subque.append

                # next target line number
                target = i + n

            # create new query and subject set pair
            qry_append(query)
            sub_append({subject})

            # update current query Id
            this = query

    # final flush
    yield qryque, subque


def range_mapper(fh, fmt=None, n=1000):
    """Read an alignment file and yield maps of query to subject(s) and their
    ranges.

    Parameters
    ----------
    fh : file handle
        Alignment file to parse.
    fmt : str, optional
        Alignment file format.
    n : int, optional
        Number of lines per chunk.

    Yields
    ------
    deque of str
        Query queue.
    deque of dict of str to list of int
        Subject-to-ranges queue.

    Notes
    -----
    Same as `plain_mapper`, except that it also returns subject ranges.

    Ranges are stored as a one-dimensional, interleaved list of start1, end1,
    start2, end2, start3, end3...

    See Also
    --------
    plain_mapper
    """
    fmt, head = (fmt, []) if fmt else infer_align_format(fh)
    parser = assign_parser(fmt, ext=True)
    qryque, subque = deque(), deque()
    qry_append, sub_append = qryque.append, subque.append
    this = None
    target = n
    for i, row in enumerate(parser(chain(iter(head), fh))):
        query, subject, _, _, start, end = row[:6]

        # range must be positive integers
        if start and end:

            if query == this:
                subque[-1].setdefault(subject, []).extend((start, end))
            else:
                if i >= target:
                    yield qryque, subque
                    qryque, subque = deque(), deque()
                    qry_append, sub_append = qryque.append, subque.append
                    target = i + n
                qry_append(query)

                # return subject Id and range
                sub_append({subject: [start, end]})

                this = query
    yield qryque, subque


def infer_align_format(fh):
    """Guess the format of an alignment file based on content.

    Parameters
    ----------
    fh : file handle
        Input alignment file.

    Returns
    -------
    str
        Alignment file format (map, b6o or sam).
    list of str
        Lines that are read in order to infer format.

    Raises
    ------
    ValueError
        Alignment file is empty or unreadable.
    ValueError
        Alignment file format cannot be determined.

    See Also
    --------
    parse_b6o_file
    parse_sam_file

    TODO
    ----
    Currently this function guesses format only based on the first line.
    This can be more robust.
    """
    # read first line of file
    try:
        line = next(fh)

    # file is empty or not readable
    except StopIteration:
        raise ValueError('Alignment file is empty or unreadable.')

    # SAM file header
    if line.split()[0] in ('@HD', '@PG'):
        return 'sam', [line]

    # tab-delimited row
    row = line.rstrip().split('\t')

    # simple query-to-subject map
    if len(row) == 2:
        return 'map', [line]

    # BLAST standard tabular format
    if len(row) >= 12:
        if all(row[i].isdigit() for i in range(3, 10)):
            return 'b6o', [line]

    # SAM format
    if len(row) >= 11:
        if all(row[i].isdigit() for i in (1, 3, 4)):
            return 'sam', [line]

    # cannot determine
    raise ValueError('Cannot determine alignment file format.')


def assign_parser(fmt, ext=False):
    """Assign parser function based on format code.

    Parameters
    ----------
    fmt : str
        Alignment file format code.
    ext : bool, optional
        Whether to get extra information.

    Returns
    -------
    function
        Alignment parser function.
    """
    if fmt == 'map':  # simple map of query <tab> subject
        return parse_map_file
    if fmt == 'b6o':  # BLAST format
        return parse_b6o_file_ext if ext else parse_b6o_file
    elif fmt == 'sam':  # SAM format
        return parse_sam_file_ext if ext else parse_sam_file
    else:
        raise ValueError(f'Invalid format code: "{fmt}".')


def parse_map_file(fh, *args):
    """Parse a simple mapping file.

    Parameters
    ----------
    fh : file handle
        Mapping file to parse.
    args : list, optional
        Placeholder for caller compatibility.

    Yields
    ------
    tuple of (str, str)
        Query and subject.

    Notes
    -----
    Only the first two columns are considered.
    """
    for line in fh:
        query, found, rest = line.partition('\t')
        if found:
            subject, found, rest = rest.partition('\t')
            yield query, subject.rstrip()


def parse_b6o_file(fh):
    """Parse a BLAST tabular file (b6o) to get basic information.

    Parameters
    ----------
    fh : file handle
        BLAST tabular file to parse.

    Yields
    ------
    tuple of (str, str)
        Query, subject.

    Notes
    -----
    BLAST tabular format:
        qseqid sseqid pident length mismatch gapopen qstart qend sstart send
        evalue bitscore

    .. _BLAST manual:
        https://www.ncbi.nlm.nih.gov/books/NBK279684/
    """
    for line in fh:
        try:
            qseqid, sseqid, _ = line.split('\t', 2)
        except ValueError:
            continue
        else:
            yield qseqid, sseqid


def parse_b6o_file_ext(fh):
    """Parse a BLAST tabular file (b6o) to get extra information.

    Parameters
    ----------
    fh : file handle
        BLAST tabular file to parse.

    Yields
    ------
    tuple of (str, str, float, int, int, int)
        Query, subject, score, length, start, end.
    """
    for line in fh:
        x = line.split('\t')
        try:
            qseqid, sseqid, length, score = x[0], x[1], int(x[3]), float(x[11])
        except IndexError:
            continue
        sstart, send = sorted((int(x[8]), int(x[9])))
        yield qseqid, sseqid, score, length, sstart, send


def parse_sam_file(fh):
    """Parse a SAM file (sam) to get basic information.

    Parameters
    ----------
    fh : file handle
        SAM file to parse.

    Yields
    ------
    tuple of (str, str)
        Query, subject.

    Notes
    -----
    SAM format:
        QNAME, FLAG, RNAME, POS, MAPQ, CIGAR, RNEXT, PNEXT, TLEN, SEQ, QUAL,
        TAGS

    .. _Wikipedia:
        https://en.wikipedia.org/wiki/SAM_(file_format)
    .. _SAM format specification:
        https://samtools.github.io/hts-specs/SAMv1.pdf
    .. _Bowtie2 manual:
        http://bowtie-bio.sourceforge.net/bowtie2/manual.shtml#sam-output

    This is by far the fastest solution. Several other solutions were tested,
    including readlines(chunk), csv.reader, and pd_read_csv(chunk). They were
    slower.
    """
    last = None

    # parse head
    for line in fh:

        # currently, it just skips head
        if line[0] != '@':
            last = line
            break

    # include current line
    it = chain([last], fh) if last else fh

    # parse body
    for line in it:

        # relevant fields
        qname, flag, rname, _ = line.split('\t', 3)

        # skip unmapped
        if rname == '*':
            continue

        # append strand to read Id if not already
        if qname[-2:] not in ('/1', '/2'):
            flag = int(flag)

            # forward strand: bit 64
            if flag & (1 << 6):
                qname += '/1'

            # reverse strand: bit 128
            elif flag & (1 << 7):
                qname += '/2'

        yield qname, rname


def parse_sam_file_ext(fh):
    """Parse a SAM file (sam) to get extra information.

    Parameters
    ----------
    fh : file handle
        SAM file to parse.

    Yields
    ------
    tuple of (str, str, None, int, int, int)
        Query, subject, None, length, start, end.
    """
    last = None
    for line in fh:
        if line[0] != '@':
            last = line
            break

    it = chain([last], fh) if last else fh
    for line in it:

        # relevant fields
        qname, flag, rname, pos, _, cigar, _ = line.split('\t', 6)
        if rname == '*':
            continue

        # leftmost mapping position
        pos = int(pos)

        # parse CIGAR string
        length, offset = cigar_to_lens(cigar)

        if qname[-2:] not in ('/1', '/2'):
            flag = int(flag)
            if flag & (1 << 6):
                qname += '/1'
            elif flag & (1 << 7):
                qname += '/2'

        yield qname, rname, None, length, pos, pos + offset - 1


@lru_cache(maxsize=128)
def cigar_to_lens(cigar):
    """Extract lengths from a CIGAR string.

    Parameters
    ----------
    cigar : str
        CIGAR string.

    Returns
    -------
    int
        Alignment length.
    int
        Offset in subject sequence.

    Notes
    -----
    This function significantly benefits from LRU cache because high-frequency
    CIGAR strings (e.g., "150M") are common and redundant calculations can be
    saved.
    """
    align, offset = 0, 0
    n = ''  # current step size
    for c in cigar:
        if c in 'MDIHNPSX=':
            if c in 'M=X':
                align += int(n)
            elif c in 'DN':
                offset += int(n)
            n = ''
        else:
            n += c
    return align, align + offset


def cigar_to_lens_ord(cigar):
    """Extract lengths from a CIGAR string.

    Parameters
    ----------
    cigar : str
        CIGAR string.

    Returns
    -------
    int
        Alignment length.
    int
        Offset in subject sequence.

    Notes
    -----
    This is an alternative solution based on unicode numbers of characters.
    It is slower than the currently adopted solution. But since it is purely
    based on numbers, there may be room for future optimization.
    """
    align, offset = 0, 0
    n = 0
    for c in map(ord, cigar):
        if c < 60:
            n = n * 10 + c - 48
        else:
            if c in (77, 88, 61):
                align += n
            elif c in (68, 78):
                offset += n
            n = 0
    return align, align + offset


def parse_kraken(fh):
    """Parse a Kraken mapping file.

    Parameters
    ----------
    fh : file handle
        Kraken mapping file to parse.

    Yields
    ------
    tuple of (str, str)
        Query and subject.

    Notes
    -----
    Kraken2 output format:
        C/U, sequence ID, taxonomy ID, length, LCA mapping

    .. _Kraken2 manual:
        https://ccb.jhu.edu/software/kraken2/index.shtml?t=manual
    """
    for line in fh:
        x = line.split('\t')
        if x[0] == 'C':
            yield x[1], x[2]


def parse_centrifuge(fh):
    """Parse a Centrifuge mapping file.

    Parameters
    ----------
    fh : file handle
        Centrifuge mapping file to parse.

    Yields
    ------
    tuple of (str, str, int, int)
        Query, subject, score, length.

    Notes
    -----
    Centrifuge output format:
        readID, seqID, taxID, score, 2ndBestScore, hitLength, queryLength,
        numMatches

    .. _Centrifuge manual:
        https://ccb.jhu.edu/software/centrifuge/manual.shtml
    """
    for line in fh:
        if not line.startswith('readID'):
            x = line.split('\t')
            yield x[0], x[1], int(x[3]), int(x[5])


def parse_sam_file_pd(fh, n=65536):
    """Parse a SAM file (sam) using Pandas.

    Parameters
    ----------
    fh : file handle
        SAM file to parse.
    n : int, optional
        Chunk size.

    Yields
    ------
    None

    Notes
    -----
    This is a SAM file parser using Pandas. It is slower than the current
    parser. The `read_csv` is fast, but the data frame manipulation slows
    down the process. It is here for reference only.
    """
    return
    # with pd.read_csv(fp, sep='\t',
    #                  header=None,
    #                  comment='@',
    #                  na_values='*',
    #                  usecols=[0, 1, 2, 3, 5],
    #                  names=['qname', 'flag', 'rname', 'pos', 'cigar'],
    #                  dtype={'qname': str,
    #                         'flag':  np.uint16,
    #                         'rname': str,
    #                         'pos':   int,
    #                         'cigar': str},
    #                  chunksize=n) as reader:
    #     for chunk in reader:
    #         chunk.dropna(subset=['rname'], inplace=True)
    #         # this is slow, because of function all
    #         chunk['length'], offset = zip(*chunk['cigar'].apply(
    #             cigar_to_lens))
    #         chunk['right'] = chunk['pos'] + offset - 1
    #         # this is slow, because of function all
    #         # chunk['qname'] = chunk[['qname', 'flag']].apply(
    #         #   qname_by_flag, axis=1)
    #         # this is a faster method
    #         chunk['qname'] += np.where(
    #             chunk['qname'].str[-2:].isin(('/1', '/2')), '',
    #             np.where(np.bitwise_and(chunk['flag'], (1 << 6)), '/1',
    #                      np.where(np.bitwise_and(chunk['flag'], (1 << 7)),
    #                      '/2', '')))
    #         chunk['score'] = 0
    #         yield from chunk[['qname', 'rname', 'score', 'length',
    #                           'pos', 'right']].values



#!/usr/bin/env python3

# ----------------------------------------------------------------------------
# Copyright (c) 2020--, Qiyun Zhu.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------


"""Functions for calculating coverage of subject sequences.

The "coverage" refers to the ranges of a subject sequence (e.g., genome) that
are covered by at least one query sequence (e.g., short sequencing read).

The functions `merge_ranges` and `parse_ranges` were modified based on the
`SortedRangeList` class implemented by Daniel Hakim (@dhakim87) in the Zebra
Filter program (https://github.com/ucsd-cmi/zebra_filter).
"""

from os import makedirs
from os.path import join


def merge_ranges(ranges):
    """Merge short, fragmental ranges into long, continuous ones.

    Parameters
    ----------
    ranges : list of int
        Ranges to merge.

    Returns
    -------
    list of int
        Merged ranges.

    Notes
    -----
    Ranges that have overlaps will be merged into one. For example:

    >>> merge_ranges([1, 3, 2, 4, 6, 8, 7, 9])
    [1, 4, 6, 9]
    """
    res = []
    res_extend = res.extend
    cstart, cend = None, None
    for start, end in sorted(zip(*[iter(ranges)] * 2)):
        if cend is None:
            # case 1: no active range, start active range
            cstart, cend = start, end
        elif cend >= start - 1:
            # case 2: active range continues through this range
            # extend active range
            cend = max(cend, end)
        else:
            # case 3: active range ends before this range begins
            # write new range out, then start new active range
            res_extend((cstart, cend))
            cstart, cend = start, end
    if cend is not None:
        res_extend((cstart, cend))
    return res


def parse_ranges(rmaps, covers, chunk=20000):
    """Extract range information from read maps.

    Parameters
    ----------
    rmaps : dict of dict
        Read map data structure.
    covers : dict of list
        Subject coverage data structure.
    chunk : int, optional
        Auto-compress after adding this many new ranges.

    Notes
    -----
    The `covers` data structure is a dictionary, where the key is a tuple of
    (sample, subject), and the value is a list. The first element of the list
    is an integer indicating how many new ranges have been added since last
    merge, and the remaining elements are individual ranges.

    Once the range counter reaches the threshold specified by `chunk`, an
    operation will be automatically triggered to merge small ranges into large,
    continuous ones, so as to reduce memory space and to improve downstream
    operations.

    See Also
    --------
    merge_ranges
    """
    add_cover = covers.setdefault
    for sample, (_, subque) in rmaps.items():
        for subjects in subque:
            for subject, ranges in subjects.items():
                cover = add_cover((sample, subject), [0, []])
                count = cover[0] + len(ranges)
                if count >= chunk:
                    cover[0] = 0
                    cover[1] = merge_ranges(cover[1] + ranges)
                else:
                    cover[0] = count
                    cover[1].extend(ranges)


def calc_coverage(covers):
    """Calculate coverage per sample per subject based on ranges.

    Parameters
    ----------
    covers : dict of list
        Subject coverage data structure.

    Returns
    -------
    dict of dict of list of int
        Subject coverage data.

    Notes
    -----
    This function is executed at last after all ranges have been added. It
    performs a final merge. Then it converts the dictionary with keys as
    (sample, subject) into a nested dictionary of sample -> subject -> ranges.

    See Also
    --------
    merge_ranges
    """
    res = {}
    add_sample = res.setdefault
    for (sample, subject), (_, ranges) in covers.items():
        add_sample(sample, {})[subject] = merge_ranges(ranges)
    return res


def write_coverage(covers, outdir):
    """Write subject coverage to per sample output files.

    Parameters
    ----------
    covers : dict of dict
        Coverage data structure.
    outdir : str
        Directory of output files.
    """
    makedirs(outdir, exist_ok=True)
    for sample, cover in sorted(covers.items()):
        with open(join(outdir, f'{sample}.cov'), 'w') as fh:
            for subject, ranges in sorted(cover.items()):
                for start, end in sorted(zip(*[iter(ranges)] * 2)):
                    print(subject, start, end, sep='\t', file=fh)



#!/usr/bin/env python3

# ----------------------------------------------------------------------------
# Copyright (c) 2020--, Qiyun Zhu.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

from unittest import TestCase, main
from os.path import join, dirname, realpath
from shutil import rmtree
from tempfile import mkdtemp

from woltka.classify import (
    assign_none, assign_free, assign_rank, counter, counter_size,
    counter_strat, counter_size_strat, majority)


class ClassifyTests(TestCase):
    def setUp(self):
        self.tmpdir = mkdtemp()
        self.datdir = join(dirname(realpath(__file__)), 'data')

    def tearDown(self):
        rmtree(self.tmpdir)

    def test_assign_none(self):
        # assign to self
        self.assertEqual(assign_none({'G1'}), 'G1')

        # count subjects
        obs = assign_none({'G1', 'G2'})
        exp = ['G1', 'G2']
        self.assertListEqual(sorted(obs), exp)

        # cannot assign
        obs = assign_none({'G1', 'G2', 'G3'}, uniq=True)
        self.assertIsNone(obs)

    def test_assign_free(self):
        tree = {'G1': 'T1', 'G2': 'T1', 'G3': 'T2',
                'T1': 'T0', 'T2': 'T0', 'T0': 'T0'}
        kwargs = {'tree': tree, 'root': 'T0'}

        # free-rank assignment
        obs = assign_free({'G1', 'G2'}, **kwargs)
        self.assertEqual(obs, 'T1')

        # root -> None
        obs = assign_free({'G1', 'G2', 'G3'}, **kwargs)
        self.assertIsNone(obs)

        # assign one sub to a taxon
        obs = assign_free({'G1'}, **kwargs)
        self.assertEqual(obs, 'T1')

        # assign one sub to itself
        obs = assign_free({'G1'}, **kwargs, subok=True)
        self.assertEqual(obs, 'G1')

        # assign one sub to nothing
        obs = assign_free({'Gx'}, **kwargs)
        self.assertIsNone(obs)

    def test_assign_rank(self):
        tree = {'G1': 'T1', 'G2': 'T1', 'G3': 'T2',
                'T1': 'T0', 'T2': 'T0', 'T0': 'T0'}
        rankdic = {'T1': 'general', 'T2': 'general', 'T0': 'marshal'}
        kwargs = {'tree': tree, 'rankdic': rankdic, 'root': 'T0'}

        # fixed-rank assignment
        obs = assign_rank({'G1', 'G2', 'G3'}, 'marshal', **kwargs)
        self.assertEqual(obs, 'T0')
        obs = assign_rank({'G1', 'G2'}, 'general', **kwargs)
        self.assertEqual(obs, 'T1')
        obs = assign_rank({'G1'}, 'general', **kwargs)
        self.assertEqual(obs, 'T1')

        # rank that does not exist
        obs = assign_rank({'G1', 'G2', 'G3'}, 'admiral', **kwargs, uniq=True)
        self.assertIsNone(obs)

        # cannot find consensus at rank
        obs = assign_rank({'G1', 'G2', 'G3'}, 'general', **kwargs, uniq=True)
        self.assertIsNone(obs)

        # majority rule
        obs = assign_rank({'G1', 'G2', 'G3'}, 'general', **kwargs, major=0.6)
        self.assertEqual(obs, 'T1')

        # consider ambiguity
        obs = assign_rank({'G1', 'G2', 'G3'}, 'general', **kwargs)
        exp = ['T1', 'T1', 'T2']
        self.assertListEqual(sorted(obs), exp)

        # assign above rank
        obs = assign_rank({'G1', 'G2', 'G3'}, 'general', tree=tree,
                          rankdic=rankdic, root=None, above=True)
        self.assertEqual(obs, 'T0')

        # assign above rank but there is a None
        obs = assign_rank({'G1', 'G2', 'G3', None}, 'general', tree=tree,
                          rankdic=rankdic, root=None, above=True)
        self.assertEqual(obs, None)

    def test_counter(self):
        taxque = ['Ecoli', ['Cdiff', 'Strep'], ['Strep', 'Ecoli', 'Cdiff',
                  'Ecoli', None], 'Ecoli', None]
        obs = counter(taxque)
        exp = {'Ecoli': 2.5, 'Cdiff': 0.75, 'Strep': 0.75}
        self.assertDictEqual(dict(obs), exp)

    def test_counter_size(self):
        # G1, G4, G6: Ecoli
        # G2, G5: Cdiff
        # G3: Strep
        # G7: none
        subque = [['G1'],
                  ['G2', 'G3'],
                  ['G3', 'G4', 'G5', 'G6', 'G7'],
                  ['G4', 'G6'],
                  ['G7']]
        taxque = ['Ecoli',
                  ['Cdiff', 'Strep'],
                  ['Strep', 'Ecoli', 'Cdiff', 'Ecoli', None],
                  'Ecoli',
                  None]
        sizes = {'G1': 5, 'G2': 5, 'G3': 6, 'G4': 9, 'G5': 2, 'G6': 4}
        obs = counter_size(subque, taxque, sizes)
        exp = {'Ecoli': 14.75,
               'Cdiff': 3.0,
               'Strep': 4.5}
        self.assertDictEqual(dict(obs), exp)
        del(sizes['G3'])
        with self.assertRaises(KeyError):
            counter_size(subque, taxque, sizes)

    def test_counter_strat(self):
        strata = {'seq1': 'Ecoli',
                  'seq2': 'Ecoli',
                  'seq3': 'Cdiff',
                  'seq4': 'Strep',
                  'seq5': 'Strep',
                  'seq6': 'Ecoli'}
        qryque = ['seq1',
                  'seq2',
                  'seq3',
                  'seq4',
                  'seq5',
                  'seq6',
                  'seq9']
        taxque = ['ligase',
                  ['polymerase', 'nuclease'],
                  ['nuclease', 'ligase', None],
                  'ligase',
                  None,
                  'nuclease',
                  ['polymerase', 'ligase']]
        obs = counter_strat(qryque, taxque, strata)
        exp = {('Ecoli', 'ligase'):     1,
               ('Ecoli', 'polymerase'): 0.5,
               ('Ecoli', 'nuclease'):   1.5,
               ('Cdiff', 'nuclease'):   0.5,
               ('Cdiff', 'ligase'):     0.5,
               ('Strep', 'ligase'):     1}
        self.assertDictEqual(dict(obs), exp)

    def test_counter_size_strat(self):
        strata = {'seq1': 'Ecoli',
                  'seq2': 'Ecoli',
                  'seq3': 'Cdiff',
                  'seq4': 'Strep',
                  'seq5': 'Strep',
                  'seq6': 'Ecoli'}
        qryque = ['seq1',
                  'seq2',
                  'seq3',
                  'seq4',
                  'seq5',
                  'seq6',
                  'seq9']
        subque = [['G1'],
                  ['G2', 'G3'],
                  ['G3', 'G4', 'G5'],
                  ['G4', 'G6'],
                  ['G7'],
                  ['G3'],
                  ['G2', 'G6']]
        taxque = ['ligase',
                  ['polymerase', 'nuclease'],
                  ['nuclease', 'ligase', None],
                  'ligase',
                  None,
                  'nuclease',
                  ['polymerase', 'ligase']]
        sizes = {'G1': 5, 'G2': 5, 'G3': 6, 'G4': 9, 'G5': 2, 'G6': 4}
        obs = counter_size_strat(qryque, subque, taxque, sizes, strata)
        exp = {('Ecoli', 'ligase'):     5,
               ('Ecoli', 'polymerase'): 2.5,
               ('Ecoli', 'nuclease'):   9,
               ('Cdiff', 'nuclease'):   3,
               ('Cdiff', 'ligase'):     4.5,
               ('Strep', 'ligase'):     6.5}
        self.assertDictEqual(dict(obs), exp)
        del(sizes['G3'])
        with self.assertRaises(KeyError):
            counter_size(subque, taxque, sizes)

    def test_majority(self):
        obs = majority([1, 1, 1, 2, 2], th=0.6)
        self.assertEqual(obs, 1)
        obs = majority([1, 1, 1, 2, 2], th=0.8)
        self.assertIsNone(obs)
        obs = majority([1, 2, 3])
        self.assertIsNone(obs)
        obs = majority([])
        self.assertIsNone(obs)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3

# ----------------------------------------------------------------------------
# Copyright (c) 2020--, Qiyun Zhu.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

"""Functions for operating BIOM tables.
"""

from functools import partial
import numpy as np
import biom
from .__init__ import __name__, __version__
from .util import rounder


def table_to_biom(data, observations, samples, metadata=None):
    """Convert table components into a BIOM table.

    Parameters
    ----------
    data : list of list
        Table data.
    observations : list
        Observation IDs.
    samples : list
        Sample IDs.
    metadata : list of dict, optional
        Observation metadata.

    Returns
    -------
    biom.Table
        Converted BIOM table.
    """
    return biom.Table(np.array(data), observations, samples, metadata or None)


def biom_to_table(table: biom.Table):
    """Convert a BIOM table into table components.

    Parameters
    ----------
    table : biom.Table
        BIOM table to convert.

    Returns
    -------
    tuple of
        list of list
            Profile data.
        list
            Observation IDs.
        list
            Sample IDs.
        list of dict
            Observation metadata.
    """
    return (table.matrix_data.toarray().astype(int).tolist(),
            table.ids('observation').tolist(),
            table.ids('sample').tolist(),
            list(map(dict, table.metadata(axis='observation') or ())))


def write_biom(table: biom.Table, fp: str):
    """Write a BIOM table to file.

    Parameters
    ----------
    table : biom.Table
        BIOM table to write.
    fp : str
        Output filepath.

    Notes
    -----
    The `generated_by` attribute of the output BIOM table will be like
    "woltka-version".
    """
    with biom.util.biom_open(fp, 'w') as f:
        table.to_hdf5(f, f'{__name__}-{__version__}')


def biom_max_f(table: biom.Table):
    """Get the maximum number of digits after the decimal place in a table.

    Parameters
    ----------
    table : biom.Table
        BIOM table to examine.

    Returns
    -------
    int
        Maximum decimal precision.
    """
    return max([0] + [str(x)[::-1].find('.') for x in
               table.matrix_data.todense().ravel().tolist()[0]])


def divide_biom(table: biom.Table, sizes: dict):
    """Divide table values by feature sizes in place.

    Parameters
    ----------
    table : biom.Table
        BIOM table to divide.
    sizes : dict
        Feature size dictionary.
    """
    table.transform(lambda data, id_, md: data / sizes[id_],
                    axis='observation', inplace=True)


def scale_biom(table: biom.Table, scale: float):
    """Multiply table values by a constant factor in place.

    Parameters
    ----------
    table : biom.Table
        BIOM table to multiply.
    scale : float
        Scale factor.
    """
    table.transform(lambda data, id_, md: data * scale, inplace=True)


def filter_biom(table: biom.Table, th: float):
    """Filter a BIOM table by per-sample count or percentage threshold.

    Parameters
    ----------
    table : biom.Table
        BIOM table to filter.
    th : float
        Per-sample minimum abundance threshold. If >= 1, this value is an
        absolute count; if < 1, it is a fraction of sum of counts.

    Returns
    -------
    biom.Table
        Filtered BIOM table.
    """
    def f(data, id_, md):
        bound = th if th >= 1 else data.sum() * th
        data[data < bound] = 0
        return data

    res = table.copy()
    res.transform(f, axis='sample')
    res.remove_empty(axis='observation')
    return res


def round_biom(table: biom.Table, digits=None):
    """Round a BIOM table's data and drop empty observations in place.

    Parameters
    ----------
    table : biom.Table
        BIOM table to round.
    digits : int, optional
        Digits after the decimal point.

    Notes
    -----
    This function will not drop empty samples.
    """
    f = np.vectorize(partial(rounder, digits=digits))
    table.transform(lambda data, id_, md: f(data), axis='observation')
    table.remove_empty(axis='observation')


def biom_add_metacol(table: biom.Table, dic, name, missing=''):
    """Add a metadata column to a table in place based on a dictionary.

    Parameters
    ----------
    table : biom.Table
        Table to add metadata column.
    dict : dict
        Metadata column (feature-to-value mapping).
    name : str
        Metadata column name.
    missing : any type, optional
        Default value if not found in dictionary.
    """
    metadata = {x: {name: dic.get(x, missing)} for x in table.ids(
        'observation')}
    table.add_metadata(metadata, axis='observation')


def collapse_biom(table: biom.Table, mapping: dict, divide=False, field=None):
    """Collapse a BIOM table in many-to-many mode.

    Parameters
    ----------
    table : biom.Table
        Table to collapse.
    mapping : dict of list of str
        Source-to-target(s) mapping.
    divide : bool, optional
        Whether divide per-target counts by number of targets per source.
    field : int, optional
        Index of field to be collapsed in a stratified table.

    Returns
    -------
    biom.Table
        Collapsed BIOM table.

    Raises
    ------
    ValueError
        Field index is not present in a feature ID.

    Notes
    -----
    Metadata will not be retained in the collapsed table.

    See Also
    --------
    .table.collapse_table
    """
    # generate metadata
    metadata = {}
    for id_ in table.ids('observation'):
        feature = id_
        if field:
            fields = feature.split('|')
            try:
                feature = fields[field]
            except IndexError:
                raise ValueError(
                    f'Feature "{feature}" has less than {field + 1} fields.')
        if feature not in mapping:
            continue
        targets = []
        for target in mapping[feature]:
            if field:
                fields[field] = target
                target = '|'.join(fields)
            targets.append(target)
        metadata[id_] = dict(part=targets)

    # filter table features
    table = table.filter(lambda data, id_, md: id_ in metadata,
                         axis='observation', inplace=False)

    # stop if no feature left
    if table.is_empty():
        return table

    # add mapping to table metadata
    table.add_metadata(metadata, axis='observation')

    # determine collapsing method
    kwargs = dict(norm=False, one_to_many=True, axis='observation',
                  one_to_many_mode=('divide' if divide else 'add'))

    # collapse table in many-to-many mode
    table = table.collapse(lambda _, md: zip(md['part'], md['part']), **kwargs)

    # round to integers
    if divide:
        round_biom(table)

    # clean up
    table.del_metadata(keys=['Path'])
    return table



#!/usr/bin/env python3

# ----------------------------------------------------------------------------
# Copyright (c) 2020--, Qiyun Zhu.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

from unittest import TestCase, main
from os.path import join, dirname, realpath
from shutil import rmtree
from tempfile import mkdtemp
from io import StringIO

from woltka.file import openzip
from woltka.align import (
    plain_mapper, range_mapper, infer_align_format, assign_parser,
    parse_map_file, parse_b6o_file, parse_b6o_file_ext, parse_sam_file,
    parse_sam_file_ext, cigar_to_lens, cigar_to_lens_ord, parse_kraken,
    parse_centrifuge, parse_sam_file_pd)


class AlignTests(TestCase):
    def setUp(self):
        self.tmpdir = mkdtemp()
        self.datdir = join(dirname(realpath(__file__)), 'data')

    def tearDown(self):
        rmtree(self.tmpdir)

    def test_plain_mapper(self):

        def _res2lst(res):
            return tuple(tuple(list(x) for x in y) for y in res)

        # simple case
        aln = StringIO('\n'.join((
            'R1	G1',
            'R2	G1',
            'R2	G2',
            'R3	G1',
            'R3	G3',
            'R4	G4',
            'R5	G5')))
        obs = plain_mapper(aln)
        exp = ((['R1', 'R2', 'R3', 'R4', 'R5'],
                [{'G1'}, {'G1', 'G2'}, {'G1', 'G3'}, {'G4'}, {'G5'}]),)
        self.assertTupleEqual(_res2lst(obs), exp)

        # chunk of 1
        aln.seek(0)
        obs = plain_mapper(aln, n=1)
        exp = ((['R1'], [{'G1'}]),
               (['R2'], [{'G1', 'G2'}]),
               (['R3'], [{'G1', 'G3'}]),
               (['R4'], [{'G4'}]),
               (['R5'], [{'G5'}]))
        self.assertTupleEqual(_res2lst(obs), exp)

        # chunk of 2
        aln.seek(0)
        obs = plain_mapper(aln, n=2)
        exp = ((['R1', 'R2'], [{'G1'}, {'G1', 'G2'}]),
               (['R3'], [{'G1', 'G3'}]),
               (['R4', 'R5'], [{'G4'}, {'G5'}]))
        self.assertTupleEqual(_res2lst(obs), exp)

        # chunk of 3
        aln.seek(0)
        obs = plain_mapper(aln, n=3)
        exp = ((['R1', 'R2'], [{'G1'}, {'G1', 'G2'}]),
               (['R3', 'R4'], [{'G1', 'G3'}, {'G4'}]),
               (['R5'], [{'G5'}]))
        self.assertTupleEqual(_res2lst(obs), exp)

        # chunk of 4
        aln.seek(0)
        obs = plain_mapper(aln, n=4)
        exp = ((['R1', 'R2', 'R3'], [{'G1'}, {'G1', 'G2'}, {'G1', 'G3'}]),
               (['R4', 'R5'], [{'G4'}, {'G5'}]))
        self.assertTupleEqual(_res2lst(obs), exp)

        # chunk of 5
        aln.seek(0)
        obs = plain_mapper(aln, n=5)
        exp = ((['R1', 'R2', 'R3'], [{'G1'}, {'G1', 'G2'}, {'G1', 'G3'}]),
               (['R4', 'R5'], [{'G4'}, {'G5'}]))
        self.assertTupleEqual(_res2lst(obs), exp)

        # format is given
        aln.seek(0)
        obs = plain_mapper(aln, fmt='map', n=5)
        self.assertTupleEqual(_res2lst(obs), exp)

        # empty alignment
        with self.assertRaises(ValueError) as ctx:
            list(plain_mapper(StringIO()))
        self.assertEqual(str(ctx.exception), (
            'Alignment file is empty or unreadable.'))

        # bad alignment
        with self.assertRaises(ValueError) as ctx:
            list(plain_mapper(StringIO('Hi there!')))
        self.assertEqual(str(ctx.exception), (
            'Cannot determine alignment file format.'))

    def test_range_mapper(self):

        def _res2lst(res):
            return tuple(tuple(list(x) for x in y) for y in res)

        aln = StringIO('\n'.join((
            'R1	G1	95	20	0	0	1	20	10	29	1	1',
            'R2	G1	95	20	0	0	1	20	16	35	1	1',
            'R3	G2	95	20	0	0	1	20	39	21	1	1',
            'R4	G2	95	20	0	0	20	1	41	22	1	1',
            'R5	G3	95	20	0	0	20	1	30	49	1	1',
            'R5	G3	95	20	0	0	20	1	50	69	1	1',
            'Rx	Gx	95	20	0	0	1	20	0	0	1	1',
            '# this is not an alignment')))
        obs = _res2lst(range_mapper(aln))[0]
        exp = [('R1', {'G1': [10, 29]}),
               ('R2', {'G1': [16, 35]}),
               ('R3', {'G2': [21, 39]}),
               ('R4', {'G2': [22, 41]}),
               ('R5', {'G3': [30, 49, 50, 69]})]
        self.assertListEqual(list(obs[0]), [x[0] for x in exp])
        self.assertListEqual(list(obs[1]), [x[1] for x in exp])

    def test_infer_align_format(self):
        # simple cases
        # map
        line = 'S1/1	NC_123456'
        obs = infer_align_format(StringIO(line))
        self.assertEqual(obs[0], 'map')
        self.assertListEqual(obs[1], [line])

        # b6o
        line = 'S1/1	NC_123456	100	100	0	0	1	100	25	124	0.1	100'
        self.assertEqual(infer_align_format(StringIO(line))[0], 'b6o')

        # sam
        line = 'S1	77	NC_123456	26	0	100M	*	0	0	*	*'
        self.assertEqual(infer_align_format(StringIO(line))[0], 'sam')

        # sam header
        line = '@HD	VN:1.0	SO:unsorted'
        self.assertEqual(infer_align_format(StringIO(line))[0], 'sam')

        # empty file
        with self.assertRaises(ValueError) as ctx:
            infer_align_format(StringIO())
        self.assertEqual(str(ctx.exception), (
            'Alignment file is empty or unreadable.'))

        # invalid sam
        line = 'S1	*	*	*	*	*	*	0	0	*	*'
        with self.assertRaises(ValueError) as ctx:
            infer_align_format(StringIO(line))
        self.assertEqual(str(ctx.exception), (
            'Cannot determine alignment file format.'))

        # cannot determine
        line = 'Hi there!'
        with self.assertRaises(ValueError) as ctx:
            infer_align_format(StringIO(line))
        self.assertEqual(str(ctx.exception), (
            'Cannot determine alignment file format.'))

        # real files
        # Bowtie2 (sam)
        with openzip(join(self.datdir, 'align', 'bowtie2', 'S01.sam.xz')) as f:
            self.assertEqual(infer_align_format(f)[0], 'sam')

        # BURST (b6o)
        with openzip(join(self.datdir, 'align', 'burst', 'S01.b6.bz2')) as f:
            self.assertEqual(infer_align_format(f)[0], 'b6o')

    def test_assign_parser(self):
        self.assertEqual(assign_parser('map'), parse_map_file)
        self.assertEqual(assign_parser('sam'), parse_sam_file)
        self.assertEqual(assign_parser('b6o'), parse_b6o_file)
        self.assertEqual(assign_parser('map', True), parse_map_file)
        self.assertEqual(assign_parser('sam', True), parse_sam_file_ext)
        self.assertEqual(assign_parser('b6o', True), parse_b6o_file_ext)
        with self.assertRaises(ValueError) as ctx:
            assign_parser('xyz')
        self.assertEqual(str(ctx.exception), (
            'Invalid format code: "xyz".'))

    def test_parse_map_file(self):
        aln = (
            'R1	G1',          # 1st line (normal)
            'R2	G2	# note',  # 2nd line (with additional column)
            '# end of file')  # 3rd line (not tab-delimited)
        obs = list(parse_map_file(aln))
        self.assertEqual(len(obs), 2)
        self.assertTupleEqual(obs[0], ('R1', 'G1'))
        self.assertTupleEqual(obs[1], ('R2', 'G2'))

    def test_parse_b6o_file(self):
        b6o = (
            '# BLAST result:',
            'S1/1	NC_123456	100	100	0	0	1	100	225	324	1.2e-30	345',
            'S1/2	NC_123456	95	98	2	1	2	99	708	608	3.4e-20	270')
        obs = list(parse_b6o_file(b6o))
        self.assertEqual(len(obs), 2)
        exp = [('S1/1', 'NC_123456'),
               ('S1/2', 'NC_123456')]
        self.assertTupleEqual(obs[0], exp[0])
        self.assertTupleEqual(obs[1], exp[1])

    def test_parse_b6o_file_ext(self):
        b6o = (
            '# BLAST result:',
            'S1/1	NC_123456	100	100	0	0	1	100	225	324	1.2e-30	345',
            'S1/2	NC_123456	95	98	2	1	2	99	708	608	3.4e-20	270')
        obs = list(parse_b6o_file_ext(b6o))
        self.assertEqual(len(obs), 2)
        exp = [('S1/1', 'NC_123456', 345, 100, 225, 324),
               ('S1/2', 'NC_123456', 270, 98, 608, 708)]
        self.assertTupleEqual(obs[0], exp[0])
        self.assertTupleEqual(obs[1], exp[1])

    def test_parse_sam_file(self):
        sam = iter((
            '@HD	VN:1.0	SO:unsorted',
            'S1	77	NC_123456	26	0	100M	*	0	0	*	*',
            'S1	141	NC_123456	151	0	80M	*	0	0	*	*',
            'S2	0	NC_789012	186	0	50M5I20M5D20M	*	0	0	*	*',
            'S2	16	*	0	0	*	*	0	0	*	*'))
        obs = list(parse_sam_file(sam))
        self.assertEqual(len(obs), 3)
        exp = [('S1/1', 'NC_123456'),
               ('S1/2', 'NC_123456'),
               ('S2',   'NC_789012')]
        # 1st line: header
        # 2nd line: normal, fully-aligned, forward strand
        self.assertTupleEqual(obs[0], exp[0])
        # 3rd line: shortened, reverse strand
        self.assertTupleEqual(obs[1], exp[1])
        # 4th line: not perfectly aligned, unpaired
        self.assertTupleEqual(obs[2], exp[2])
        # 5th line: not aligned

        # header only
        sam = iter((
            '@HD	VN:1.0	SO:unsorted',
            '@SQ	SN:G000005825	LN:4249288',
            '@SQ	SN:G000006175	LN:1936387'))
        obs = list(parse_sam_file(sam))
        self.assertEqual(len(obs), 0)

        # file is empty
        obs = list(parse_sam_file(iter(())))
        self.assertEqual(len(obs), 0)

    def test_parse_sam_file_ext(self):
        sam = iter((
            '@HD	VN:1.0	SO:unsorted',
            'S1	77	NC_123456	26	0	100M	*	0	0	*	*',
            'S1	141	NC_123456	151	0	80M	*	0	0	*	*',
            'S2	0	NC_789012	186	0	50M5I20M5D20M	*	0	0	*	*',
            'S2	16	*	0	0	*	*	0	0	*	*'))
        obs = list(parse_sam_file_ext(sam))
        self.assertEqual(len(obs), 3)
        exp = [('S1/1', 'NC_123456', None, 100, 26, 125),
               ('S1/2', 'NC_123456', None, 80, 151, 230),
               ('S2', 'NC_789012', None, 90, 186, 280)]
        self.assertTupleEqual(obs[0], exp[0])
        self.assertTupleEqual(obs[1], exp[1])
        self.assertTupleEqual(obs[2], exp[2])

        obs = list(parse_sam_file_ext(iter(())))
        self.assertEqual(len(obs), 0)

    def test_cigar_to_lens(self):
        self.assertTupleEqual(cigar_to_lens('150M'), (150, 150))
        self.assertTupleEqual(cigar_to_lens('3M1I3M1D5M'), (11, 12))

    def test_cigar_to_lens_ord(self):
        self.assertTupleEqual(cigar_to_lens_ord('150M'), (150, 150))
        self.assertTupleEqual(cigar_to_lens_ord('3M1I3M1D5M'), (11, 12))

    def test_parse_kraken(self):
        kra = ('C	S1	561	150	561:100 A:10 562:40',
               'U	S2	0	150	1:80 A:40 0:20 A:10')
        obs = list(parse_kraken(kra))
        self.assertEqual(len(obs), 1)
        exp = [('S1', '561')]
        self.assertTupleEqual(obs[0], exp[0])

    def test_parse_centrifuge(self):
        cen = ('readID	seqID	taxID	score	2ndBestScore	'
               'hitLength	queryLength	numMatches',
               'S1	NC_123456	561	125	0	50	150	1')
        obs = list(parse_centrifuge(cen))
        self.assertEqual(len(obs), 1)
        exp = [('S1', 'NC_123456', 125, 50)]
        self.assertTupleEqual(obs[0], exp[0])

    def test_parse_sam_file_pd(self):
        self.assertIsNone(parse_sam_file_pd([]))


if __name__ == '__main__':
    main()
